import numpy as np
import librosa, cv2
from scipy import signal
import ruptures as rpt
from functools import partial
import matplotlib.pyplot as plt
from .external import PyOctaveBand


########################################################################################################################################
## RECIPES (Combination of Transformations)
########################################################################################################################################

################################################################
## RecipeGeneral
################################################################
class RecipeGeneral(object):
    '''
    소음 분석에 사용하는 일반적인 1D, 2D feature들을 계산함 - 자동 분석에 사용
    (현재) octave band pressure (1/3), rms, cog, sbw, flat, zcr, ose, stft, mel,
    '''
    
    def __init__(self, sr, specs=None):
        '''
        params
          sr
          specs
          steady <list or tuple>   range of steady [start_sec, end_sec]
        '''
        self.sr = sr
        self.specs = {'height': 224, 'hop_sec': 0.01, 'fft_sec': 0.08, 'unit': 'amplitude'} if specs is None else specs        
        
        self.hpf = BandPassFilter(sr)
        self.features = {
            # PCM
            'pcm': lambda y, sr: y,
            # RMS
            'rms': RMS(sr, **{k: v for k, v in self.specs.items() if k in ['hop_sec', 'fft_sec', 'unit']}),
            # Center of Gravity
            'cog': SpectralCentroid(sr, **{k: v for k, v in self.specs.items() if k in ['hop_sec', 'fft_sec']}),
            # Spectral Bandwidth
            'sbw': SpectralBandwidth(sr, **{k: v for k, v in self.specs.items() if k in ['hop_sec', 'fft_sec']}),
            # Spectral Flatness
            'flat': SpectralFlatness(sr, **{k: v for k, v in self.specs.items() if k in ['hop_sec', 'fft_sec']}),
            # Zero Crossing Rate
            'zcr': ZeroCrossingRate(sr, **{k: v for k, v in self.specs.items() if k in ['hop_sec', 'fft_sec']}),
            # Onset Strength Envelope
            'ose': OnsetStrength(sr, **{k: v for k, v in self.specs.items() if k in ['height', 'hop_sec', 'fft_sec']}),
            # STFT
            'stft': STFT(sr, **self.specs),
            # Mel Spectrogram
            'mel': MelSpectrogram(sr, **self.specs),
            # Octave Band Sound Level (pressure)
            'obp': OctaveBandFilter(sr),
        }
        
    def __call__(self, y, sr):
        y = self.hpf(y, sr)
        features = {k: v(y, sr) for k, v in self.features.items()}
        
        return features


################################################################
## RecipeMel
################################################################
class RecipeMel(object):
    '''
    1. Fixed hop length, 서로 다른 3개의 n_fft를 갖는 mel spectrograms을 구하여 image 형식 (heigth x width x 3 ch, 8 bit depth)의 array 생성
    2. image 형식의 array를 augment (image와 동일하게 처리) 
      (1) square crop (width만 crop하여 square image를 구함)
      (2) roll
      (3) flip
    
    return
      3d numpy array (height x width x channel)
    
    init. arguments
      height: default is 224 pixel (= number of frequency bins) to use ImageNet CNN models.
      hop_sec: default 0.01 sec (= 10 ms), hop_length = sampling rate * hop_sec
      n_fft_factors: n_fft_factor * hop_length = n_fft
    '''
    def __init__(self, height=224, hop_sec=0.01, fft_secs=[0.08, 0.08, 0.08]):
        self.specs = []
        for fft_sec in fft_secs:
            self.specs.append(MelSpectrogram(height=height, hop_sec=hop_sec, fft_sec=fft_sec))
        
    def __call__(self, y, sr, augment=False):
        '''
        input
          y [numpy array], sr [int]
          
        return
          S [dB]
        '''
        # Stack Channels
        channels = []
        for spec in self.specs:
            S = spec(y, sr)
            S = ((S-S.min())/(S.max()-S.min()) * 255).astype(np.int)
            channels.append(S)
        x = np.stack(channels, -1)        
        
        # Augmentation
        if augment:
            x = SquareCrop('random')(x)
            x = Roll()(x)
            x = Flip()(x)
        else:
            x = SquareCrop('left')(x)
        
        return x

################################################################
## RecipeSTFT ~ multi layer
################################################################
class RecipeSTFT(object):
    '''
    1. 동일한 hop length와 서로 다른 3개의 n_fft를 갖는 STFT를 구하여 image 형식 (heigth x width x 3 ch, 8 bit depth)의 array 생성
    2. image 형식의 array를 augment (image와 동일하게 처리) 
      (1) square crop (width만 crop하여 square image를 구함)
      (2) roll
      (3) flip
    
    return 
      3d numpy array (height x width x channel).
    
    init. arguments
      height: default is 224 pixel (= number of frequency bins)
      hop_sec: default 0.01 sec (= 10 ms), hop_length = sampling rate * hop_sec
      n_fft_factors: n_fft_factor * hop_length = n_fft
    '''
    def __init__(self, height=224, hop_sec=0.01, fft_secs=[0.08, 0.16, 0.32]):
        self.specs = []
        for fft_sec in fft_secs:
            self.specs.append(STFT(height=height, hop_sec=hop_sec, fft_sec=fft_sec))
        
    def __call__(self, y, sr, augment=False):
        '''
        input
          y [numpy array], sr [int]
          
        return
          S [dB]
        '''
        # Stack STFT spec.
        channels = []
        for spec in self.specs:
            S = spec(y, sr)
            S = ((S-S.min())/(S.max()-S.min()) * 255).astype(np.int)
            channels.append(S)
        x = np.stack(channels, -1)
        
        # Augmentation
        if augment:
            x = SquareCrop('random')(x)
            x = Roll()(x)
            x = Flip()(x)
        else:
            x = SquareCrop('left')(x)
        
        return x


##########################################################################################################################################
## FEATURE TRANSFORMS - FILTER, TEMPORAL, SPECTRAL, SPECTROGRAM, ...
##########################################################################################################################################

################################################################
## BandPassFilter (low-pass, band-pass, high-pass)
################################################################
class BandPassFilter(object):
    '''
    Bandpass filter이지만 cutoofs의 left가 None이면 lowpass, right가 None이면 highpass로 사용할 수 있음.
    
    init.
    :params sr:
    :params cutoffs:
    :parmas order:
    
    call
    :params y: data
    :params sr: sampling rate
    
    retur
    filtered y
    '''
    def __init__(self, sr=None, cutoffs=[12, None], order=5, ):
        # check cutoff
        for co in cutoffs:
            if (co is not None) and (sr is not None):
                if co > sr/2:
                    raise ValueError("[ERROR] cutoff frequency is higher than Nyquist limit.")
        
        # filter coefficient
        if (cutoffs[0] is not None) and (cutoffs[1] is None):
            sos = signal.butter(N=order, Wn=cutoffs[0], btype='highpass', fs=sr, output='sos')
        elif (cutoffs[0] is None) and (cutoffs[1] is not None):
            sos = signal.butter(N=order, Wn=cutoffs[1], btype='lowpass', fs=sr, output='sos')
        elif (cutoffs[0] is not None) and (cutoffs[1] is not None):
            sos = signal.butter(N=order, Wn=cutoffs, btype='bandpass', fs=sr, output='sos')
        else:
            raise ValueError("[ERROR] cutoffs required; cutoffs=[cultoff_left, cutoff_right]")
        
        self.sr = sr
        self.cutoffs = cutoffs
        self.order = order 
        self.sos = sos
        
    def __call__(self, y, sr=None):
        # re-initiate if sr is changed
        if sr != self.sr:
            print(f"WARNING: input sr ({sr}) is not equal to initiated sr ({self.sr}), instance is reinitiated w/ sr {sr}")
            self.__init__(sr=sr, cutoffs=self.cutoffs, order=self.order)
            
        return signal.sosfiltfilt(self.sos, y)


################################################################
# estimate changepoints
################################################################
def estimate_changepoints(X, use_diff=True, verbose=False):
    '''
    params
      X          array (n x x)
      use_diff   np.diff 후 changepoint 추정 (default)
    return
      bkpts <list>, states <np.array>, steady <int>
    '''
    
    # model & search range
    model = rpt.Pelt(model='rbf', jump=10)
    pens = range(1, 50)
    
    # use_diff
    X_ = np.pad(np.diff(X), ((0,0),(0,1)), 'edge') if use_diff else X

    # 빈도 가장 높은 no. of breaks를 선택
    l, n, n_max = 0, 0, 0
    bkpts = [X_.shape[1]]
    for pen in pens:
        b = model.fit_predict(X_.T, pen, )
        l_b = len(b)
        
        if l_b < 2:
            break
        elif l_b == l:
            n += 1
            if n > n_max:
                bkpts = b
                n_max = n
        else:
            n = 1
            l = l_b

    bkpt_prev = 0
    states = []
    for lab, bkpt in enumerate(bkpts):
        states += [lab] * (bkpt - bkpt_prev)
        bkpt_prev = bkpt
        
    # to np.array
    states = np.array(states, dtype=np.int)
    
    # estimate steady
    std_min = np.Inf
    for s in sorted(set(states)):
        std = X[:, states==s].std()
        if std < std_min:
            steady = s
            std_min = std
    
    if verbose:
        print(f"Check Changepoints:")
        print(f"  - {len(bkpts)} sections estimated, 0-{'-'.join([f'(section {str(i)})-{str(x)}' for i, x in enumerate(bkpts)])}")
        
    # last bkpt will be ignored - 닫힌 구간 표시용임
    return bkpts, states, steady


################################################################
## [SPECTRAL] Octave Band Sound Level
################################################################
# 나중에 OctaveBandLevel 등으로 이름 바꿀 것
class OctaveBandFilter(object):
    '''
    Octave band filter
    
    init.
    :param sr:
    :param fraction: bandwidth 'b', 1/3-octave b=3, 2/3-octave b=3/2, 1-octave b=1 (default 3)
    :param order:
    :param limits: 
    
    call
    :param y: y
    :param sr: sampling rate
    
    return 
    spl[dB]
    '''
    def __init__(self, sr=None, fraction=3, order=6, fmin=12, fmax=12800):
        
        # estimate sr if sr not specified
        self.sr = sr if sr else fmax * 2
        self.fraction = fraction
        self.order = order
        self.fmin = fmin
        self.fmax = fmax
        
        # generate frequencies & remove outer frequency
        freq_arr = PyOctaveBand.getansifrequencies(self.fraction, self.fmin, self.fmax)
        idx = np.where(freq_arr[:, 2] > self.sr / 2)
        if any(idx[0]):
            freq_arr = np.delete(freq_arr, idx, axis=0)
        self.freqs = freq_arr[:, 0]
        self.freq_limits = freq_arr[:, 1:]
        self.N = self.freqs.shape[0]
        
        # Calculate the downsampling factor (array of integers with size [freq])   
        guard = 0.10
        factor = np.floor((self.sr / (2 + guard)) / self.freq_limits[:, 1])
        factor = np.clip(factor, 1, 50).astype('int')
        self.factor = factor
        
        # Get SOS filter coefficients (3d array w/ shape (freq, order, 6))
        butter = partial(signal.butter, self.order, btype='bandpass', output='sos')
        wns = self.freq_limits / (self.sr / self.factor /2).reshape(-1, 1)
        sos = np.apply_along_axis(butter, 1, wns)
        self.sos = sos
        
    def __call__(self, x, sr):
        
        if sr != self.sr:
            print(f"WARNING: input sr ({sr}) is not equal to initiated sr ({self.sr}), instance is reinitiated w/ sr {sr}")
            self.__init__(sr, self.fraction, self.order, self.fmin, self.fmax)
            
        spl = np.zeros(self.N)
            
        for idx in range(self.N):
            y = signal.decimate(x, self.factor[idx])
            y = signal.sosfilt(self.sos[idx], y)
            y = np.std(y)
            spl[idx] = 20 * np.log10(y / 2e-5)
        
        return spl

    
################################################################
## FFT filter (arithmetic filter)
################################################################
# class FFT(object):
#     '''
    
#     '''
#      def __init__(self, sr=None, freq_width=60, harmonic_only=False):
        
#         if not sr:
#             self.sr = None
#             self.sr_ = 
#             self.freq_width = freq_width
#             self.harmonic_only = harmonic_only
        
#         else:         
#             # estimate sr if sr not specified
#             self.sr = sr if sr else fmax * 2
#             self.fraction = fraction
#             self.order = order
#             self.fmin = fmin
#             self.fmax = fmax

#             # generate frequencies & remove outer frequency
#             freq_arr = PyOctaveBand.getansifrequencies(self.fraction, self.fmin, self.fmax)
#             idx = np.where(freq_arr[:, 2] > self.sr / 2)
#             if any(idx[0]):
#                 freq_arr = np.delete(freq_arr, idx, axis=0)
#             self.freqs = freq_arr[:, 0]
#             self.freq_limits = freq_arr[:, 1:]
#             self.N = self.freqs.shape[0]

#             # Calculate the downsampling factor (array of integers with size [freq])   
#             guard = 0.10
#             factor = np.floor((self.sr / (2 + guard)) / self.freq_limits[:, 1])
#             factor = np.clip(factor, 1, 50).astype('int')
#             self.factor = factor

#             # Get SOS filter coefficients (3d array w/ shape (freq, order, 6))
#             butter = partial(signal.butter, self.order, btype='bandpass', output='sos')
#             wns = self.freq_limits / (self.sr / self.factor /2).reshape(-1, 1)
#             sos = np.apply_along_axis(butter, 1, wns)
#             self.sos = sos
        
#     def __call__(self, x, sr):
        
#         if sr != self.sr:
#             print(f"WARNING: sr ({sr}) is not equal to current sr ({self.sr}), sr updated.")
#             self.__init__(sr, self.fraction, self.order, self.fmin, self.fmax)
            
#         spl = np.zeros(self.N)
            
#         for idx in range(self.N):
#             y = signal.decimate(x, self.factor[idx])
#             y = signal.sosfilt(self.sos[idx], y)
#             y = np.std(y)
#             spl[idx] = 20 * np.log10(y / 2e-5)
        
#         return spl


################################################################
## [TIME-DOMAIN] RMS
################################################################
class RMS(object):
    '''
    root mean squared spl along time axis - 각 time bin에서 energy의 RMS
    '''
    def __init__(self, sr, hop_sec=0.01, fft_sec=0.08, center=True, unit='amplitude'):
        self.sr = sr
        self.hop_sec = hop_sec
        self.fft_sec = fft_sec
        self.center = center
        self.unit = unit.lower()
        
        self.frame_length = int(self.sr * self.fft_sec)
        self.hop_length = int(self.sr * self.hop_sec)
        
        if self.unit in ['db']:
            self.unit_convert = lambda x: 20 * np.log10(x)
        elif self.unit in ['amp', 'amplitude']:
            self.unit_convert = lambda x: x
        elif self.unit in ['power']:
            self.unit_convert = lambda x: x**2
        
    def __call__(self, y, sr=None):
        '''
        params
          y
          sr  for API consistency (not using)
        '''
        rms = librosa.feature.rms(y, frame_length=self.frame_length, hop_length=self.hop_length, center=self.center)
        if y.ndim == 1:
            rms = rms[0]
        rms = self.unit_convert(rms)
        
        return rms
    
    
################################################################
## [TIME-DOMAIN] SpectralCentroid (Center of Gravity)
################################################################
class SpectralCentroid(object):
    '''
    spectral centroid along time axis - 각 time bin에서 frequency의 center
    '''
    def __init__(self, sr, hop_sec=0.01, fft_sec=0.08, center=True):
        self.sr = sr
        self.hop_sec = hop_sec
        self.fft_sec = fft_sec
        self.center = center
        
        self.n_fft = int(self.sr * self.fft_sec)
        self.hop_length = int(self.sr * self.hop_sec)
        
    def __call__(self, y, sr=None):
        '''
        params
          y
          sr
        '''
        if sr is None:
            sr = self.sr
            
        cog = librosa.feature.spectral_centroid(y, sr, n_fft=self.n_fft, hop_length=self.hop_length, center=self.center)
        if y.ndim == 1:
            cog = cog[0]
        
        return cog
    
    
################################################################
## [TIME-DOMAIN] SpectralBandwidth
################################################################
class SpectralBandwidth(object):
    '''
    spectral bandwidth along time axis - 각 time bin에서 frequency의 bandwidth
    '''
    def __init__(self, sr, hop_sec=0.01, fft_sec=0.08, p=2, center=True):
        self.sr = sr
        self.hop_sec = hop_sec
        self.fft_sec = fft_sec
        self.p = p
        self.center = center
        
        self.n_fft = int(self.sr * self.fft_sec)
        self.hop_length = int(self.sr * self.hop_sec)
        
    def __call__(self, y, sr=None):
        '''
        params
          y
          sr
        '''
        if sr is None:
            sr = self.sr
            
        sbw = librosa.feature.spectral_bandwidth(y, sr, n_fft=self.n_fft, hop_length=self.hop_length, center=self.center, p=self.p)
        if y.ndim == 1:
            sbw = sbw[0]
        
        return sbw
     
    
################################################################
## [TEMPORAL] SpectralFlatness
################################################################
class SpectralFlatness(object):
    '''
    spectral flatness along time axis - 각 time bin에서 frequency의 bandwidth
    '''
    def __init__(self, sr, hop_sec=0.01, fft_sec=0.08, center=True, unit='amplitude'):
        self.sr = sr
        self.hop_sec = hop_sec
        self.fft_sec = fft_sec
        self.center = center
        self.unit = unit
        
        self.n_fft = int(self.sr * self.fft_sec)
        self.hop_length = int(self.sr * self.hop_sec)
        
        if self.unit in ['db']:
            self.power = 1
            self.unit_convert = lambda x: 20 * np.log10(x)
        elif self.unit in ['amp', 'amplitude']:
            self.power = 1
            self.unit_convert = lambda x: x
        elif self.unit in ['power']:
            self.power = 2
            self.unit_convert = lambda x: x
        
    def __call__(self, y, sr=None):
        '''
        params
          y
          sr  for API consistency (not using)
        '''
            
        flat = librosa.feature.spectral_flatness(y, n_fft=self.n_fft, hop_length=self.hop_length, center=self.center, power=self.power)
        if y.ndim == 1:
            flat = flat[0]
        flat = self.unit_convert(flat)
        
        return flat
    
    
################################################################
## [TEMPORAL] ZeroCrossingRate
################################################################
class ZeroCrossingRate(object):
    '''
    root mean squared spl along time axis - 각 time bin에서 energy의 RMS
    '''
    def __init__(self, sr, hop_sec=0.01, fft_sec=0.08, center=True):
        self.sr = sr
        self.hop_sec = hop_sec
        self.fft_sec = fft_sec
        self.center = center
        
        self.frame_length = int(self.sr * self.fft_sec)
        self.hop_length = int(self.sr * self.hop_sec)
        
    def __call__(self, y, sr=None):
        '''
        params
          y
          sr  for API consistency (not using)
        '''
        zcr = librosa.feature.zero_crossing_rate(y, frame_length=self.frame_length, hop_length=self.hop_length, center=self.center)
        if y.ndim == 1:
            zcr = zcr[0]
        
        return zcr
        
    
################################################################
## [Onset] OnsetStrength
################################################################
class OnsetStrength(object):
    '''
    onset strength
    (fixed) mel feature, power=1.0 
    '''
    def __init__(self, sr, height=224, hop_sec=0.01, fft_sec=0.08, fmin=0, fmax=None, center=True):
        self.sr = sr
        self.height = height
        self.hop_sec = hop_sec
        self.fft_sec = fft_sec
        self.center = center
        self.fmin = fmin
        fmax = fmax if fmax is not None else int(sr/2)
        self.fmax = fmax
        
        self.n_fft = int(self.sr * self.fft_sec)
        self.hop_length = int(self.sr * self.hop_sec)
        
    def __call__(self, y, sr=None):
        '''
        params
          y
          sr
        '''
        if sr is None:
            sr = self.sr
            
        ose = librosa.onset.onset_strength(
            y, sr, n_fft=self.n_fft, hop_length=self.hop_length, n_mels=self.height, fmin=self.fmin, fmax=self.fmax, center=self.center, 
        )
        
        return ose

    
################################################################
## Mel Spectrogram
################################################################
class MelSpectrogram(object):
    '''
    librosa.feature.melspectrogram을 height, hop_sec에 대해 다시 작성한 것.
    
    return 
      2d numpy array, unit dB
    
    init. arguments
      height: default is 224 pixel (= number of frequency bins)
      hop_sec: default 0.01 sec (= 10 ms), hop_length = sampling rate * hop_sec    
    '''
    def __init__(self, sr=None, height=224, hop_sec=0.01, fft_sec=0.08, fmin=0, fmax=None, center=True, unit='dB'):
        self.sr = sr
        self.height = height
        self.hop_sec = hop_sec
        self.fft_sec = fft_sec
        self.fmin = fmin
        fmax = fmax if fmax is not None else int(sr/2)
        self.fmax = fmax
        self.center = True
        
        # calcs
        self.hop_length = int(sr * self.hop_sec)
        self.n_fft = int(sr * self.fft_sec)
        self.freqs = librosa.mel_frequencies(n_mels=height, fmin=fmin, fmax=fmax)
        
        if unit.lower() in ['db']:
            self.power = 2
            # self.unit_convert = librosa.power_to_db
            self.unit_convert = lambda x: 10 * np.log10(x)
        elif unit.lower() in ['amp', 'amplitude']:
            self.power = 1
            self.unit_convert = lambda x: x
        elif unit.lower() in ['power']:
            self.power = 2
            self.unit_convert = lambda x: x
            
        # placeholder
        self.S = None
        
    def __call__(self, y, sr):
        # check sr
        if sr != self.sr:
            print(f"WARNING: input sr ({sr}) is not equal to initiated sr ({self.sr}), instance is reinitiated w/ sr {sr}")
            self.__init__(sr, self.hight, self.hop_sec, self.fft_sec, self.output_unit)
            
        # get mel
        S = librosa.feature.melspectrogram(
            y, sr, 
            n_fft=self.n_fft, 
            hop_length=self.hop_length, 
            n_mels=self.height, 
            fmin=self.fmin, 
            fmax=self.fmax, 
            center=self.center, 
            power=self.power
        )
        
        # unit convert
        S = self.unit_convert(S)
        
        # update S
        self.S = S
        
        return S
    
    def show(self, savepath=None, H=None, dpi=72, ticks=False):
        if self.S is None:
            return
        
        S = self.S
        
        if H is None:
            H = S.shape[0] / dpi
            
        W = H / S.shape[0] * S.shape[1]
        
        # plot
        fig, ax = plt.subplots(figsize=(W, H))
        ax.imshow(S, origin='lower', cmap='magma', )
        
        # ticks
        if ticks:    
            plt.xticks(
                [i for i in np.arange(0, S.shape[1]+1) if i % int(1/self.hop_sec) == 0], 
                [int(i*self.hop_sec) for i in np.arange(0, S.shape[1]+1) if i % int(1/self.hop_sec) == 0], 
            )
            plt.yticks(
                [i for i, e in enumerate(self.freqs) if i % int(S.shape[0]/10) == 0], 
                [np.int(e) for i, e in enumerate(self.freqs) if i % int(S.shape[0]/10) == 0], 
            )
        else:
            ax.set_axis_off()
        
        # save
        if savepath is not None:
            plt.savefig(savepath, dpi=dpi)
        else:
            plt.show()

    
################################################################
## STFT
################################################################
class STFT(object):
    '''
    librosa.STFT를 height, hop_sec에 대해 다시 작성한 것.
    계산 중 sr이 변경되면 self.sr re-initiate됨.
    height는 cv2로 resize되고, height=None이면 resize 없이 n_fft에 따라 height 결정됨
    
    return 
      2d numpy array, unit dB
    
    init. arguments
      height: default is 224 pixel (= number of frequency bins)
      hop_sec: default 0.01 sec (= 10 ms), hop_length = sampling rate * hop_sec
      
    example
      get_stft = STFT()
      S = get_stft(y, sr)
    '''
    def __init__(self, sr, height=224, hop_sec=0.01, fft_sec=None, center=True, unit='dB'):
        
        self.sr = sr
        self.height = height
        self.hop_sec = hop_sec
        self.fft_sec = fft_sec if fft_sec is not None else hop_sec * 8
        self.center = center
        self.unit = unit.lower()
        
        # calcs
        self.hop_length = int(sr * hop_sec)
        self.n_fft = int(sr * fft_sec)
        self.freqs = librosa.fft_frequencies(sr, self.n_fft)
        
        if self.unit in ['db']:
            # self.unit_convert = lambda x: librosa.amplitude_to_db(np.abs(x))
            self.unit_convert = lambda x: 20 * np.log10(np.abs(x))
        elif self.unit in ['amp', 'amplitude']:
            self.unit_convert = lambda x: np.abs(x)
        elif self.unit in ['power']:
            self.unit_convert = lambda x: np.abs(x) ** 2
        elif self.unit in ['complex']:
            self.unit_convert = lambda x: x
            if self.height is not None:
                print('WARNING!!! if unit is complex, resize cannot be done. height is changed to None.')
        else:
            print('WARNING!!! wrong unit, it will return dB')
            self.unit_convert = lambda x: librosa.amplitude_to_db(np.abs(x))
        
        # avoid cv2 freeze issue
        cv2.setNumThreads(0)
        
    def __call__(self, y, sr):

        if sr != self.sr:
            print(f"WARNING: input sr ({sr}) is not equal to initiated sr ({self.sr}), instance is reinitiated w/ sr {sr}")
            self.__init__(sr, height=self.hight, hop_sec=self.hop_sec, fft_sec=self.fft_sec, center=self.center, unit=self.unit)
        
        # get stft
        S = librosa.stft(
            y, 
            n_fft=self.n_fft, 
            hop_length=self.hop_length,
            center=self.center,
        )
            
        # unit convert
        S = self.unit_convert(S)
        
        # resize
        if self.height is not None:
            S = cv2.resize(S, (S.shape[1], self.height))
        
        return S
    
    def show(self, savepath=None, H=None, dpi=72, ticks=False):
        if self.S is None:
            return
        
        S = self.S
        
        if H is None:
            H = S.shape[0] / dpi
            
        W = H / S.shape[0] * S.shape[1]
        
        # plot
        fig, ax = plt.subplots(figsize=(W, H))
        ax.imshow(S, origin='lower', cmap='magma', )
        
        # ticks
        if ticks:    
            plt.xticks(
                [i for i in np.arange(0, S.shape[1]+1) if i % int(1/self.hop_sec) == 0], 
                [int(i*self.hop_sec) for i in np.arange(0, S.shape[1]+1) if i % int(1/self.hop_sec) == 0], 
            )
            plt.yticks(
                [i for i, e in enumerate(self.freqs) if i % int(S.shape[0]/10) == 0], 
                [np.int(e) for i, e in enumerate(self.freqs) if i % int(S.shape[0]/10) == 0], 
            )
        else:
            ax.set_axis_off()
        
        # save
        if savepath is not None:
            plt.savefig(savepath, dpi=dpi)
        else:
            plt.show()
  
    
    
  ################################################################
  ## DWT
  ################################################################



  ################################################################
  ## CWT
  ################################################################



  ################################################################
  ## Scale
  ################################################################
  # 0 ~ 1 float
  # 0 ~ 255 uint8
class FreqStandardization(object):
    '''
    Frequency Standardization (Mean, std are required)
    
    init arguments
      mean, std
      
    return
      input
    '''
    def __init__(self, mean=None, std=None, cutoff=(-3, 3)):
        self.mean = mean
        self.std = std
        self.cutoff = cutoff
        
    def __call__(self, y, sr):
        '''
        input
          y [numpy array], sr [int]
          
        return
          S [dB]
        '''
        if all([x is not None for x in [self.mean, self.std]]):
            y = (y - self.mean) / self.std
            cutL, cutH = self.cutoff
            y = np.clip(y, cutL, cutH)
            y = (y - cutL) / (cutH - cutL) 
        else:
            print("mean, std are required!")
        return y, sr


########################################################################################################################################
## IMAGE TRANSFORMATION (Crop, Flip, Rolling, ...)
########################################################################################################################################

################################################################
#### Square Crop
################################################################
class ToImage(object):
    '''
    Spectrogram stack하여 이미지로 변환, 8 bit로 convert
    (추가할 것 ~ normalize, standardize)
    '''
    def __init__(self):
        pass
    
    def __call__(self, channels:list):
        img = []
        for ch in channels:
            # normalize 방법에 따라 바뀔 수 있도록 할 것
            ch = ((ch-ch.min())/(ch.max()-ch.min()) * 255).astype(np.int)
            img.append(ch)           
        return np.stack(img, -1)

    
################################################################
## Square Crop
################################################################
class SquareCrop(object):   
    '''
    모든 변환은 Channel Last 기준 (h x w x c), pytorch에서 사용하려면 (c x h )
    '''
    def __init__(self, position='left'):
        self.position = position
    
    def __call__(self, x):
        '''
        input
          2d numpy array w/ shape (h x w)
          
        return
          2d numpy array w/ shape (h x h)
        '''
        h, w, c = x.shape
            
        #### if width is smaller than height, drop the data 
        if w <= h:
            return x

        #### do crop
        if self.position == 'left':
            x = x[:, :h, :]
        elif self.position == 'center': 
            x = x[:, w//2-h//2:w//2+h//2, :]
        elif self.position == 'right':
            x = x[:, -h:, :]
        elif self.position == 'random':
            l = np.random.randint(0, w - h)
            x = x[:, l:l+h, :]
        else:
            print('ERROR!!! crop options: [left, center, right, random]')  
                
        return x

    
################################################################
## Resize
################################################################
class Resize(object):
    '''
    모든 변환은 Channel Last 기준 (HWC)
    '''
    def __init__(self, height=224, width='fixed_aspect'):
        self.height = height
        self.width = width
        cv2.setNumThreads(0)
        
    def __call__(self, x):
        
        #### REORDER
        h, w, c = x.shape
        
        #### SET WIDTH
        if self.width == 'fixed_aspect':
            width = int(w * self.height / h)
        elif self.width == 'fixed_width':
            width = w
        elif type(width) is int:
            width = width
        else:
            print('ERROR!!! resize options: "fixed_aspect", "fixed_width", int()')
        
        #### RESIZE
        x = cv2.resize(x, (width, self.height))
        
        return x
    
    
################################################################
## Rolling
################################################################
class Roll(object):
    '''
    모든 변환은 Channel Last 기준 (HWC)
    '''
    def __init__(self, shift='random', freq=0.3):
        self.shift = shift
        self.freq = freq
    
    def __call__(self, x):
        
        if np.random.random() < self.freq:
            
            #### Get Order
            axis = 1   # H: 0, W: 1, C: 2
            width = x.shape[axis]
            
            #### ROLL
            if self.shift == 'random':
                shift = np.random.randint(0, width)
            elif type(self.shift) is int:
                shift = self.shift
            else:
                print("ERROR!!! only 'random' and integer are available now")
            x = np.roll(x, shift, axis)
            
        return x
        
        
################################################################
## Flip
################################################################
class Flip(object):
    '''
    모든 변환은 Channel Last 기준 (HWC)
    '''
    def __init__(self, freq=0.3):
        self.freq = freq
    
    def __call__(self, x):
        
        if np.random.random() < self.freq:
            #### Flip (mirror)
            x = cv2.flip(x, 1)

        return x
    

################################################################################################################################
# ETC. - API consistency 만족하지 않는 기타 함수
################################################################################################################################

